/* $Id: screen.c,v 1.27 2010/09/13 12:44:57 lamonte Exp $
 *
 * screen.c
 *
 * Additional acknowledgements, news, and updates to the Qodem Project
 * are available at http://qodem.sourceforge.net .
 *
 * This module is licensed under the GNU General Public License Version 2.
 * Please see the file "COPYING" in this directory for more information
 * about the GNU General Public License Version 2.
 *
 *     Copyright (C) 2009  Kevin Lamonte
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	 USA
 *
 *
 */

/*
 * This module contains the ncurses-based screen drawing functions.
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#include "common.h"
#include "qcurses.h"
#include "qodem.h"
#include "input.h"
#include "colors.h"
#include "music.h"
#include "linux.h"
#include "states.h"
#include "status.h"
#include "codepage.h"
#include "screen.h"

/*
 * Turn a Q_COLOR_* enum into the color part of the ncurses attr_t.
 * This also performs the 0x38 <--> 0x00 switch.
 */
static attr_t screen_color(const int q_color) {
	attr_t color = (q_text_colors[q_color].fg << 3) | q_text_colors[q_color].bg;
	if (color == 0x38) {
		color = q_white_color_pair_num;
	} else if (color == 0x00) {
		color = 0x38;
	}
	return color;
} /* ---------------------------------------------------------------------- */

/*
 * Turn a Q_COLOR_* enum into the color part of the ncurses attr_t.
 */
static attr_t scrollback_color(const int q_color) {
	attr_t color = (q_text_colors[q_color].fg << 3) | q_text_colors[q_color].bg;
	return color;
} /* ---------------------------------------------------------------------- */

/*
 * Turn a Q_COLOR_* enum into the non-color part of the ncurses attr_t
 */
static attr_t screen_attr(const int q_color) {
	attr_t attr = A_NORMAL;
	if (q_text_colors[q_color].bold == Q_TRUE) {
		attr |= A_BOLD;
	}
	return attr;
} /* ---------------------------------------------------------------------- */

/*
 * Turn a Q_COLOR_* enum into a ncurses attr_t
 */
attr_t scrollback_full_attr(const int q_color) {
	return COLOR_PAIR(scrollback_color(q_color)) | screen_attr(q_color);
} /* ---------------------------------------------------------------------- */

attr_t color_from_attr(const attr_t color) {
	return PAIR_NUMBER(color);
} /* ---------------------------------------------------------------------- */

attr_t color_to_attr(const attr_t color) {
	return COLOR_PAIR(color);
} /* ---------------------------------------------------------------------- */

void screen_put_scrollback_char_yx(const int y, const int x, const wchar_t ch, const attr_t attr) {
	static cchar_t ncurses_ch;
	static wchar_t ch_cached = -1;
	static attr_t attr_cached = A_BLINK | A_PROTECT;	/* Something I won't use often */
	static int cache_count = 0;
	wchar_t wch[2];
	attr_t attr2;
	attr_t color;
	if ((ch == ch_cached) && (attr == attr_cached)) {
		/* NOP */
		cache_count++;
	} else {
		attr2 = attr & 0xFFFF00FFUL;
		color = PAIR_NUMBER(attr);
		if (color == 0x38) {
			color = q_white_color_pair_num;
		} else if (color == 0x00) {
			color = 0x38;
		}
		wch[0] = ch;
		wch[1] = 0;
		setcchar(&ncurses_ch, wch, attr2, color, NULL);
		ch_cached = ch;
		attr_cached = attr;
	}
	mvadd_wch(y, x, &ncurses_ch);
} /* ---------------------------------------------------------------------- */

/*
 * Put ch on screen with attr and color
 */
void screen_win_put_char(void * win, const wchar_t ch, const attr_t attr, const attr_t color) {
	cchar_t ncurses_ch;
	wchar_t wch[2];
	wch[0] = ch;
	wch[1] = 0;
	setcchar(&ncurses_ch, wch, attr, color, NULL);
	wadd_wch(win, &ncurses_ch);
} /* ---------------------------------------------------------------------- */

void screen_win_put_char_yx(void * win, const int y, const int x, const wchar_t ch, const attr_t attr, const attr_t color) {
	cchar_t ncurses_ch;
	wchar_t wch[2];
	wch[0] = ch;
	wch[1] = 0;
	setcchar(&ncurses_ch, wch, attr, color, NULL);
	mvwadd_wch(win, y, x, &ncurses_ch);
} /* ---------------------------------------------------------------------- */

void screen_win_put_str(void * win, const char * str, const attr_t attr, const attr_t color) {
#if 0
	int i;
	for (i = 0; i < strlen(str); i++) {
		screen_win_put_char(win, str[i], attr, color);
	}
#else
	int i;
	int x;
	int y;
	getyx((WINDOW *)win, y, x);
	cchar_t ncurses_ch[Q_MAX_LINE_LENGTH];
	wchar_t wch[2];
	wch[1] = 0;
	for (i = 0; i < strlen(str); i++) {
		wch[0] = str[i];
		setcchar(&ncurses_ch[i], wch, attr, color, NULL);
	}
	wch[0] = 0;
	setcchar(&ncurses_ch[strlen(str)], wch, attr, color, NULL);
	wadd_wchstr((WINDOW *)win, ncurses_ch);
	wmove((WINDOW *)win, y, x + strlen(str));
#endif
} /* ---------------------------------------------------------------------- */

void screen_win_put_wcs(void * win, const wchar_t * wcs, const attr_t attr, const attr_t color) {
#if 0
	int i;
	for (i = 0; i < wcslen(str); i++) {
		screen_win_put_char_yx(win, y, x + i, wcs[i], attr, color);
	}
#else
	int i;
	int x;
	int y;
	getyx((WINDOW *)win, y, x);
	cchar_t ncurses_ch[Q_MAX_LINE_LENGTH];
	wchar_t wch[2];
	wch[1] = 0;
	for (i = 0; i < wcslen(wcs); i++) {
		wch[0] = wcs[i];
		setcchar(&ncurses_ch[i], wch, attr, color, NULL);
	}
	wch[0] = 0;
	setcchar(&ncurses_ch[wcslen(wcs)], wch, attr, color, NULL);
	mvwadd_wchstr((WINDOW *)win, y, x, ncurses_ch);
	wmove((WINDOW *)win, y, x + wcslen(wcs));
#endif
} /* ---------------------------------------------------------------------- */

void screen_win_put_str_yx(void * win, const int y, const int x, const char * str, const attr_t attr, const attr_t color) {
#if 0
	int i;
	for (i = 0; i < strlen(str); i++) {
		screen_win_put_char_yx(win, y, x + i, str[i], attr, color);
	}
#else
	int i;
	cchar_t ncurses_ch[Q_MAX_LINE_LENGTH];
	wchar_t wch[2];
	wch[1] = 0;
	for (i = 0; i < strlen(str); i++) {
		wch[0] = str[i];
		setcchar(&ncurses_ch[i], wch, attr, color, NULL);
	}
	wch[0] = 0;
	setcchar(&ncurses_ch[strlen(str)], wch, attr, color, NULL);
	mvwadd_wchstr((WINDOW *)win, y, x, ncurses_ch);
	wmove((WINDOW *)win, y, x + strlen(str));
#endif
} /* ---------------------------------------------------------------------- */

void screen_win_put_strn(void * win, const char * str, const int n, const attr_t attr, const attr_t color) {
	int i;
	for (i = 0; (i < strlen(str)) && (i < n); i++) {
		screen_win_put_char(win, str[i], attr, color);
	}
} /* ---------------------------------------------------------------------- */

void screen_win_put_strn_yx(void * win, const int y, const int x, const char * str, const int n, const attr_t attr, const attr_t color) {
	int i;
	for (i = 0; (i < strlen(str)) && (i < n); i++) {
		screen_win_put_char_yx(win, y, x + i, str[i], attr, color);
	}
} /* ---------------------------------------------------------------------- */

void screen_win_put_wcs_yx(void * win, const int y, const int x, const wchar_t * wcs, const attr_t attr, const attr_t color) {
#if 0
	int i;
	for (i = 0; i < wcslen(str); i++) {
		screen_win_put_char_yx(win, y, x + i, wcs[i], attr, color);
	}
#else
	int i;
	cchar_t ncurses_ch[Q_MAX_LINE_LENGTH];
	wchar_t wch[2];
	wch[1] = 0;
	for (i = 0; i < wcslen(wcs); i++) {
		wch[0] = wcs[i];
		setcchar(&ncurses_ch[i], wch, attr, color, NULL);
	}
	wch[0] = 0;
	setcchar(&ncurses_ch[wcslen(wcs)], wch, attr, color, NULL);
	mvwadd_wchstr((WINDOW *)win, y, x, ncurses_ch);
	wmove((WINDOW *)win, y, x + wcslen(wcs));
#endif
} /* ---------------------------------------------------------------------- */

void screen_win_put_hline_yx(void * win, const int y, const int x, const wchar_t ch, const int n, const attr_t attr, const attr_t color) {
	cchar_t ncurses_ch;
	wchar_t wch[2];
	wch[0] = ch;
	wch[1] = 0;
	setcchar(&ncurses_ch, wch, attr, color, NULL);
	mvwhline_set((WINDOW *)win, y, x, &ncurses_ch, n);
} /* ---------------------------------------------------------------------- */

void screen_win_put_vline_yx(void * win, const int y, const int x, const wchar_t ch, const int n, const attr_t attr, const attr_t color) {
	cchar_t ncurses_ch;
	wchar_t wch[2];
	wch[0] = ch;
	wch[1] = 0;
	setcchar(&ncurses_ch, wch, attr, color, NULL);
	mvwvline_set((WINDOW *)win, y, x, &ncurses_ch, n);
} /* ---------------------------------------------------------------------- */

void screen_win_put_printf_yx(void * win, const int y, const int x, const attr_t attr, const attr_t color, const char * format, ...) {
	va_list arglist;
	char outbuf[DIALOG_MESSAGE_SIZE];
	assert(strlen(format) < DIALOG_MESSAGE_SIZE);
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsnprintf((char *)(outbuf + strlen(outbuf)), (DIALOG_MESSAGE_SIZE - strlen(outbuf)), format, arglist);
	va_end(arglist);

	screen_win_put_strn_yx((WINDOW *)win, y, x, outbuf, strlen(outbuf), attr, color);
} /* ---------------------------------------------------------------------- */

void screen_put_char_yx(const int y, const int x, const wchar_t ch, const attr_t attr, const attr_t color) {
	screen_win_put_char_yx(stdscr, y, x, ch, attr, color);
} /* ---------------------------------------------------------------------- */

void screen_put_str_yx(const int y, const int x, const char * str, const attr_t attr, const attr_t color) {
	screen_win_put_str_yx(stdscr, y, x, str, attr, color);
} /* ---------------------------------------------------------------------- */

void screen_put_printf_yx(const int y, const int x, const attr_t attr, const attr_t color, const char * format, ...) {
	va_list arglist;
	char outbuf[DIALOG_MESSAGE_SIZE];
	assert(strlen(format) < DIALOG_MESSAGE_SIZE);
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsnprintf((char *)(outbuf + strlen(outbuf)), (DIALOG_MESSAGE_SIZE - strlen(outbuf)), format, arglist);
	va_end(arglist);

	screen_win_put_strn_yx(stdscr, y, x, outbuf, strlen(outbuf), attr, color);
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_char(void * win, const wchar_t ch, const int q_color) {
	screen_win_put_char((WINDOW *)win, ch, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_char_yx(void * win, const int y, const int x, const wchar_t ch, const int q_color) {
	screen_win_put_char_yx((WINDOW *)win, y, x, ch, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_str(void * win, const char * str, const int q_color) {
	screen_win_put_str((WINDOW *)win, str, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_wcs(void * win, const wchar_t * wcs, const int q_color) {
	screen_win_put_wcs((WINDOW *)win, wcs, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_str_yx(void * win, const int y, const int x, const char * str, const int q_color) {
	screen_win_put_str_yx((WINDOW *)win, y, x, str, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_wcs_yx(void * win, const int y, const int x, const wchar_t * wcs, const int q_color) {
	screen_win_put_wcs_yx((WINDOW *)win, y, x, wcs, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_strn(void * win, const char * str, const int n, const int q_color) {
	screen_win_put_strn((WINDOW *)win, str, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_strn_yx(void * win, const int y, const int x, const char * str, const int n, const int q_color) {
	screen_win_put_strn_yx((WINDOW *)win, y, x, str, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_hline_yx(void * win, const int y, const int x, const wchar_t ch, const int n, const int q_color) {
	screen_win_put_hline_yx((WINDOW *)win, y, x, ch, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_vline_yx(void * win, const int y, const int x, const wchar_t ch, const int n, const int q_color) {
	screen_win_put_vline_yx((WINDOW *)win, y, x, ch, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_printf(void * win, const int q_color, const char * format, ...) {
	va_list arglist;
	char outbuf[DIALOG_MESSAGE_SIZE];
	assert(strlen(format) < DIALOG_MESSAGE_SIZE);
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsnprintf((char *)(outbuf + strlen(outbuf)), (DIALOG_MESSAGE_SIZE - strlen(outbuf)), format, arglist);
	va_end(arglist);

	screen_win_put_strn((WINDOW *)win, outbuf, strlen(outbuf), screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_win_put_color_printf_yx(void * win, const int y, const int x, const int q_color, const char * format, ...) {
	va_list arglist;
	char outbuf[DIALOG_MESSAGE_SIZE];
	assert(strlen(format) < DIALOG_MESSAGE_SIZE);
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsnprintf((char *)(outbuf + strlen(outbuf)), (DIALOG_MESSAGE_SIZE - strlen(outbuf)), format, arglist);
	va_end(arglist);

	screen_win_put_strn_yx((WINDOW *)win, y, x, outbuf, strlen(outbuf), screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_char(const wchar_t ch, const int q_color) {
	screen_win_put_char(stdscr, ch, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_char_yx(const int y, const int x, const wchar_t ch, const int q_color) {
	screen_win_put_char_yx(stdscr, y, x, ch, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_str(const char * str, const int q_color) {
	screen_win_put_str(stdscr, str, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_wcs(const wchar_t * wcs, const int q_color) {
	screen_win_put_wcs(stdscr, wcs, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_str_yx(const int y, const int x, const char * str, const int q_color) {
	screen_win_put_str_yx(stdscr, y, x, str, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_wcs_yx(const int y, const int x, const wchar_t * wcs, const int q_color) {
	screen_win_put_wcs_yx(stdscr, y, x, wcs, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_strn(const char * str, const int n, const int q_color) {
	screen_win_put_strn(stdscr, str, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_strn_yx(const int y, const int x, const char * str, const int n, const int q_color) {
	screen_win_put_strn_yx(stdscr, y, x, str, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_hline_yx(const int y, const int x, const wchar_t ch, const int n, const int q_color) {
	screen_win_put_hline_yx(stdscr, y, x, ch, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_vline_yx(const int y, const int x, const wchar_t ch, const int n, const int q_color) {
	screen_win_put_vline_yx(stdscr, y, x, ch, n, screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_printf(const int q_color, const char * format, ...) {
	va_list arglist;
	char outbuf[DIALOG_MESSAGE_SIZE];
	assert(strlen(format) < DIALOG_MESSAGE_SIZE);
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsnprintf((char *)(outbuf + strlen(outbuf)), (DIALOG_MESSAGE_SIZE - strlen(outbuf)), format, arglist);
	va_end(arglist);

	screen_win_put_strn(stdscr, outbuf, strlen(outbuf), screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_put_color_printf_yx(const int y, const int x, const int q_color, const char * format, ...) {
	va_list arglist;
	char outbuf[DIALOG_MESSAGE_SIZE];
	assert(strlen(format) < DIALOG_MESSAGE_SIZE);
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsnprintf((char *)(outbuf + strlen(outbuf)), (DIALOG_MESSAGE_SIZE - strlen(outbuf)), format, arglist);
	va_end(arglist);

	screen_win_put_strn_yx(stdscr, y, x, outbuf, strlen(outbuf), screen_attr(q_color), screen_color(q_color));
} /* ---------------------------------------------------------------------- */

void screen_move_yx(const int y, const int x) {
	move(y, x);
} /* ---------------------------------------------------------------------- */

void screen_win_move_yx(void * win, const int y, const int x) {
	wmove((WINDOW *)win, y, x);
} /* ---------------------------------------------------------------------- */

void screen_flush() {
	refresh();
} /* ---------------------------------------------------------------------- */

void screen_win_flush(void * win) {
	wrefresh((WINDOW *)win);
} /* ---------------------------------------------------------------------- */

void screen_clear() {
	werase(stdscr);
} /* ---------------------------------------------------------------------- */

void screen_really_clear() {
	int i;
	cchar_t ncurses_ch;
	wchar_t wch[2];
	wch[0] = ' ';
	wch[1] = 0;
	setcchar(&ncurses_ch, wch, A_NORMAL, 0x1, NULL);
	for (i = 0; i < HEIGHT; i++) {
		mvhline_set(i, 0, &ncurses_ch, WIDTH);
	}
	refresh();
} /* ---------------------------------------------------------------------- */

void screen_resize(int height, int width) {
#ifdef Q_PDCURSES
	resize_term(height, width);
#else
	resizeterm(height, width);
#endif
} /* ---------------------------------------------------------------------- */

void screen_win_get_yx(void * win, int * y, int * x) {
	int screen_x;
	int screen_y;
	getyx((WINDOW *)win, screen_y, screen_x);
	*y = screen_y;
	*x = screen_x;
} /* ---------------------------------------------------------------------- */

/* Beep support */
void screen_beep() {
	struct q_music_struct p;

	if (q_status.beeps == Q_FALSE) {
		/* Don't beep */
		return;
	}

	switch (q_status.emulation) {

	case Q_EMUL_TTY:
	case Q_EMUL_DEBUG:
	case Q_EMUL_ANSI:
	case Q_EMUL_AVATAR:
	case Q_EMUL_VT52:
	case Q_EMUL_VT100:
	case Q_EMUL_VT102:
	case Q_EMUL_VT220:
	case Q_EMUL_XTERM:
	case Q_EMUL_XTERM_UTF8:
		/* Most emulations just beep normally. */
		beep();
		break;

	case Q_EMUL_LINUX:
	case Q_EMUL_LINUX_UTF8:
		/*
		 * Linux emulation is different:  we have to beep using the
		 * correct frequency and duration.
		 */
		/* "Play" the beep */
		memset(&p, 0, sizeof(struct q_music_struct));
		p.hertz = q_linux_beep_frequency;
		p.duration = q_linux_beep_duration;
		play_music(&p);
		break;
	}

} /* ---------------------------------------------------------------------- */

void screen_setup() {
#ifdef Q_PDCURSES
#ifdef XCURSES
	char * pdcursesOptions[8] = {
		"qodem",
		"-normalFont",
		"-xos4-terminus-medium-r-normal--20-200-72-72-c-100-iso10646-1",
		"-lines",
		"35",
		"-cols",
		"80",
		0
	};
	Xinitscr(7, pdcursesOptions);
#else
	initscr();
#endif /* XCURSES */
#else
	initscr();
#endif /* Q_PDCURSES */

	getmaxyx(stdscr, HEIGHT, WIDTH);
	STATUS_HEIGHT = 1;
	/*
	 * I remember re-reading the worklog.html and wondering how I managed to
	 * get ^Z and ^C passed in.  Here it is:  curses call to enable raw mode.
	 */
	raw();
	nodelay(stdscr, TRUE);
	q_keyboard_blocks = Q_FALSE;
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	meta(stdscr, TRUE);
	keypad(stdscr, TRUE);
	start_color();
	q_setup_colors();

	/* Set color AFTER they've been initialized! */
	q_current_color = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
} /* ---------------------------------------------------------------------- */

void screen_teardown() {
	endwin();
} /* ---------------------------------------------------------------------- */

void screen_clear_remaining_line() {
	int x, y;
	int i;
	if (q_white_color_pair_num == 0) {
		/* fg white bg black */
		clrtoeol();
	} else {
		/* some other background color */
		getyx(stdscr, y, x);
		for (i = x; i < WIDTH; i++) {
			screen_put_char_yx(y, i, ' ', 0, q_white_color_pair_num);
		}
		move(y, x);
	}
} /* ---------------------------------------------------------------------- */

void screen_shell_mode() {
	reset_shell_mode();
} /* ---------------------------------------------------------------------- */

void screen_program_mode() {
	reset_prog_mode();
} /* ---------------------------------------------------------------------- */

void screen_get_dimensions(int * height, int * width) {
	int local_height;
	int local_width;
	getmaxyx(stdscr, local_height, local_width);
	*height = local_height;
	*width = local_width;
} /* ---------------------------------------------------------------------- */

void * screen_subwin(int height, int width, int top, int left) {
	return screen_win_subwin(stdscr, height, width, top, left);
} /* ---------------------------------------------------------------------- */

void * screen_win_subwin(void * win, int height, int width, int top, int left) {
	WINDOW * window = subwin(stdscr, height, width, top, left);
	meta(window, TRUE);
	keypad(window, TRUE);
	return window;
} /* ---------------------------------------------------------------------- */

void screen_delwin(void * win) {
	delwin(win);
} /* ---------------------------------------------------------------------- */

/*
 * draw_box - draw a standard-looking box
 */
void screen_draw_box(const int left, const int top, const int right, const int bottom) {
	screen_win_draw_box(stdscr, left, top, right, bottom);
} /* ---------------------------------------------------------------------- */

/*
 * draw_box - draw a standard-looking box
 */
void screen_win_draw_box(void * window, const int left, const int top, const int right, const int bottom) {
	int i;
	int window_height;
	int window_length;

	window_length = right - left;
	window_height = bottom - top;

	screen_win_put_color_char_yx(window, top, left, cp437_chars[Q_WINDOW_LEFT_TOP], Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_char_yx(window, top, left + window_length - 1, cp437_chars[Q_WINDOW_RIGHT_TOP], Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_char_yx(window, top + window_height - 1, left, cp437_chars[Q_WINDOW_LEFT_BOTTOM], Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_char_yx(window, top + window_height - 1, left + window_length - 1, cp437_chars[Q_WINDOW_RIGHT_BOTTOM], Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_hline_yx(window, top, left + 1, cp437_chars[Q_WINDOW_TOP], window_length - 2, Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_vline_yx(window, top + 1, left, cp437_chars[Q_WINDOW_SIDE], window_height - 2, Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_hline_yx(window, top + window_height - 1, left + 1, cp437_chars[Q_WINDOW_TOP], window_length - 2, Q_COLOR_WINDOW_BORDER);
	screen_win_put_color_vline_yx(window, top + 1, left + window_length - 1, cp437_chars[Q_WINDOW_SIDE], window_height - 2, Q_COLOR_WINDOW_BORDER);

	/* Ick, but... */
	for (i=1; i<window_height - 1; i++) {
		screen_win_put_color_hline_yx(window, i + top, 1 + left, ' ', window_length - 2, Q_COLOR_WINDOW);
	}

} /* ---------------------------------------------------------------------- */
